# External Imports
# Import only with "import package",
# it will make explicity in the code where it came from.

import requests
import urllib3

# Internal Imports
# Import only with "from x import y", to simplify the code.

from ..common.exceptions import InvalidOptionError
from ..common.exceptions import RESTAPIError

# Disable urllib3 SSL warnings
urllib3.disable_warnings()


def rest_format(text: str) -> str:
    """
    Converts a string to be compatible with iControl REST API.

    Arguments:
        text: String to be converted.

    """

    text = text.replace("/", "~")
    text = text.replace("%", "%25")
    return text


def token(device: str, username: str, password: str,
          login_provider: str = "tmos",
          debug: str = None) -> str:
    """
    Gets a token from the device.

    Arguments:
        device: Name or IP of the device to send the REST requests.
        username: Username used to login to the device.
        password: Password used to login to the device.
        login_provider: Login provider used to authenticate the user.
        debug: Debug file name to be used to output the debug information.

    Exceptions:
        InvalidOptionError: Raised when invalid options are used as arguments.
    """

    # Required for 12.0.0, but not in 15.1.0
    auth = (username, password)
    data = {}
    data["username"] = username
    data["password"] = password
    data["loginProviderName"] = login_provider
    response = requests.post(
        f"https://{device}/mgmt/shared/authn/login",
        json=data, verify=False, auth=auth)
    if response.status_code != 200:
        raise RESTAPIError(response, debug)
    return response.json()["token"]["token"]


def refresh_token(device: str, username: str, password: str,
                  login_provider: str = "tmos",
                  debug: str = None) -> str:
    """
    Gets a refresh token from the device.

    Arguments:
        device: Name or IP of the device to send the REST requests.
        username: Username used to login to the device.
        password: Password used to login to the device.
        login_provider: Login provider used to authenticate the user.
        debug: Debug file name to be used to output the debug information.

    Exceptions:
        InvalidOptionError: Raised when invalid options are used as arguments.
    """

    # Required for 12.0.0, but not in 15.1.0
    auth = (username, password)
    data = {}
    data["username"] = username
    data["password"] = password
    data["loginProviderName"] = login_provider
    response = requests.post(
        f"https://{device}/mgmt/shared/authn/login",
        json=data, verify=False, auth=auth)
    if response.status_code != 200:
        raise RESTAPIError(response, debug)
    response_json = response.json()
    if "refreshToken" in response_json:
        return response.json()["refreshToken"]["token"]
    else:
        raise InvalidOptionError(
            "Refresh token only available in a BIG-IQ device.")



# External Imports
# Import only with "import package",
# it will make explicity in the code where it came from.

# Turns all annotations into string literals.
# This is one exception to the external import rule.
from __future__ import annotations
import time

# Internal Imports
# Import only with "from x import y", to simplify the code.
from .big import BIG
from .common.exceptions import RESTAPIError
from .common.restobject import RESTObject


class BIGIQ(BIG):
    """
    Defines methods to call the iControl REST API that can be used by BIG-IQ.

    Arguments:
        device: Name or IP of the device to send the HTTP requests.
        username: Username used to login to the device.
        password: Password used to login to the device.
        login_provider: Login provider used to authenticate the user.
        request_token: Indicates if a token should be requested from the
            device and used for HTTP requests.
        token: Token to be used to send HTTP requests to the device.
        refresh_token: Refresh token to be used to request new token,
            the new token will be then used for HTTP requests.
        debug: Debug file name to be used to output the debug information.

    Exceptions:
        InvalidOptionError: Raised when invalid options are used as arguments.
    """

    def task_start(self, path: str, data: dict) -> RESTObject:
        """
        Starts a task on the device.

        Sends an HTTP POST request to the iControl REST API.

        Arguments:
            path: HTTP path used in the HTTP request sent to the device.
            data: Payload that will be sent to the device.

        Exceptions:
            RESTAPIError: Raised when iControl REST API returns an error.
        """

        if self.request_token or self.refresh_token is not None:
            self._check_token()
        url = self._get_url(path)
        response = self.session.post(url, json=data)
        if response.status_code != 202:
            raise RESTAPIError(response, self.debug)
        return RESTObject(response.json())

    def task_wait(self, obj: RESTObject, interval: int = 10) -> None:
        """
        Continually queries the status of the task until it finishes.

        Sends an HTTP GET request to the iControl REST API.

        Arguments:
            obj: Object that represents the task.
            interval: The interval the queries will be made.

        Exceptions:
            RESTAPIError: Raised when iControl REST API returns an error.
        """

        if self.request_token or self.refresh_token is not None:
            self._check_token()
        path = self._get_path(obj)
        url = self._get_url(path)
        while True:
            if self.request_token or self.refresh_token is not None:
                self._check_token()
            response = self.session.get(url)
            if response.status_code != 200:
                raise RESTAPIError(response, self.debug)
            status = response.json()["status"]
            if status == "FAILED":
                raise RESTAPIError(response, self.debug)
            if status == "FINISHED":
                return
            else:
                time.sleep(interval)

    def task_completed(self, obj: RESTObject) -> bool:
        """
        Verifies if the task is completed.

        Sends an HTTP GET request to the iControl REST API.

        Arguments:
            obj: Object that represents the task.

        Exceptions:
            RESTAPIError: Raised when iControl REST API returns an error.
        """

        if self.request_token or self.refresh_token is not None:
            self._check_token()
        path = self._get_path(obj)
        url = self._get_url(path)
        response = self.session.get(url)
        if response.status_code != 200:
            raise RESTAPIError(response, self.debug)
        status = response.json()["status"]
        if status == "FAILED":
            raise RESTAPIError(response, self.debug)
        if status == "FINISHED":
            return True
        else:
            return False

    def exist(self, path: str) -> bool:
        """
        Verifies if the object exists on the device.

        Sends an HTTP GET request to the iControl REST API.

        Arguments:
            path: HTTP path used in the HTTP request sent to the device.

        Exceptions:
            RESTAPIError: Raised when iControl REST API returns an error.
        """

        if self.request_token or self.refresh_token is not None:
            self._check_token()
        url = self._get_url(path)
        response = self.session.get(url)
        if response.status_code != 200:
            raise RESTAPIError(response, self.debug)
        response_json = response.json()
        if response_json["items"]:
            return True
        else:
            return False

    def link(self, path: str) -> str:
        """
        Get the link for the object.

        Sends an HTTP GET request to the iControl REST API.

        Arguments:
            path: HTTP path used in the HTTP request sent to the device.

        Exceptions:
            RESTAPIError: Raised when iControl REST API returns an error.
        """

        obj = self.load(path)[0]
        return obj.properties["selfLink"]

    def id(self, path: str) -> str:
        """
        Get the ID of the object.

        Sends an HTTP GET request to the iControl REST API.

        Arguments:
            path: HTTP path used in the HTTP request sent to the device.

        Exceptions:
            RESTAPIError: Raised when iControl REST API returns an error.
        """

        obj = self.load(path)[0]
        link = obj.properties["selfLink"]
        id_ = link.split("/")[-1]
        return id_



# External Imports
# Import only with "import package",
# it will make explicity in the code where it came from.

# Turns all annotations into string literals.
# This is one exception to the external import rule.
from __future__ import annotations
import json


class RESTObject():
    """
    Represents a object created from a response of the iControl REST API.

    Parameters:
        properties: Represents the properties of the object.
    """

    def __init__(self, properties: dict) -> RESTObject:
        self.properties = properties

    def asdict(self) -> dict:
        """Converts the object to a dictionary."""

        # iControl REST API only accepts enabled=True or disabled=True
        if "enabled" in self.properties:
            if self.properties["enabled"] is True:
                self.properties.pop("enabled")
                self.properties["disabled"] = True
        if "disabled" in self.properties:
            if self.properties["disabled"] is True:
                self.properties.pop("disabled")
                self.properties["enabled"] = True

        return self.properties

    def __str__(self) -> str:
        """Converts the object to a string."""

        return json.dumps(self.properties, indent=4)



"""Defines a list of exceptions."""

# External Imports
# Import only with "import package",
# it will make explicity in the code where it came from.

# Turns all annotations into string literals.
# This is one exception to the external import rule.
from __future__ import annotations
import requests
import json


class Error(Exception):
    """Base class for all SDK exceptions."""
    pass


class RESTAPIError(Error):
    """
    Exception for API errors from iControl REST API.

    Parameters:
        response: A object created from a response to a HTTP request.
        debug: If set, indicates the file to save the debug output created.
    """

    def __init__(self, response: requests.Response,
                 debug: str) -> RESTAPIError:
        reponse_content_type = response.headers.get("Content-Type")
        try:
            response_json = json.dumps(response.json(), indent=4)
            response_body = f"\nResponse Body:\n{response_json}"
        except Exception:
            if reponse_content_type == "application/octet-stream":
                response_body = f"\nResponse Body:\n{response.content}"
            else:
                response_body = f"\nResponse Body:\n{response.text}"
        response_status = response.status_code
        if debug is not None:
            request_method = response.request.method
            request_url = response.request.url
            request_headers = response.request.headers
            request_headers = json.dumps(dict(request_headers), indent=4)
            request_body = response.request.body
            if request_body is not None:
                try:
                    request_body = json.dumps(
                        dict(eval(request_body)), indent=4)
                except Exception:
                    pass
            else:
                request_body = ""
            response_headers = response.headers
            response_headers = json.dumps(dict(response_headers), indent=4)
            information = (f"Request Method:\n{request_method}"
                           f"\nRequest URL:\n{request_url}"
                           f"\nRequest headers:\n{request_headers}"
                           f"\nRequest body:\n{request_body}"
                           f"\nResponse Status:\n{response_status}"
                           f"\nResponse headers:\n{response_headers}")
            with open(debug, "w") as file_:
                file_.write(f"{information}{response_body}")
                super().__init__(f"Debug information saved in file {debug}.")
        else:
            information = f"\nStatus:\n{response_status}"
            super().__init__(f"{information}{response_body}")


class InvalidOptionError(Error):
    """
    Exception raised for invalid options passed.

    Parameters:
        message: Message that will be printed with the exception.
    """
    def __init__(self, message: str) -> InvalidOptionError:
        super().__init__(message)
